/*==============================================================================
		初期設定ファイルアクセスクラス	for ＢｅＯＳ		Ver 1.53
		Copyright (C) 1999 by ＳＨＩＮＴＡ
		Last Modified 1999.09.04.Sa by SHINTA
==============================================================================*/

/*==============================================================================
 Ver. |					主な改良点
--------------------------------------------------------------------------------
 ---- | 1999.02.22.Mo 開発着手。
 1.00 | 1999.03.01.Mo オリジナルバージョン。最低限の機能を搭載。
 1.10 | 1999.03.02.Tu アクティブセクション方式に変更した。
 1.20 | 1999.03.02.Tu 大文字小文字の区別をしないようにした。初公開。
 1.30 | 1999.03.11.Th SectionExists()、KeyExists()を作った。
 1.40 | 1999.03.12.Fr DeleteSection()、DeleteKey()を作った。
 1.50 | 1999.03.14.Su EnumSections()、EnumKeys()を作った。
==============================================================================*/

/*=== Memo =====================================================================
==============================================================================*/

#include "IniFile.h"
//------------------------------------------------------------------------------
TIniFile::TIniFile(int iMaxLen) : TSHIFile()
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::TIniFile(const entry_ref *ref,uint32 open_mode,int iMaxLen) : TSHIFile(ref,open_mode)
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::TIniFile(const BEntry *entry,uint32 open_mode,int iMaxLen) : TSHIFile(entry,open_mode)
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::TIniFile(const char *path,uint32 open_mode,int iMaxLen) : TSHIFile(path,open_mode)
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::TIniFile(const BDirectory *dir,const char *path,uint32 open_mode,int iMaxLen)
		: TSHIFile(dir,path,open_mode)
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::TIniFile(const BFile &file,int iMaxLen) : TSHIFile(file)
{
	Init(iMaxLen);
}
//------------------------------------------------------------------------------
TIniFile::~TIniFile()
{
	Flush();
	DestroyVar();
}
//------------------------------------------------------------------------------
bool	TIniFile::ChangeSection(const char *cSecName)
{
	if ( (mxSecCurrent = FindSection(cSecName)) == NULL )
		mxSecCurrent = CreateSection(cSecName);
	return true;
}
//------------------------------------------------------------------------------
TIniFileSection	*TIniFile::CreateSection(const char *cSecName,int iIdx)
{
	TIniFileSection	*xSection;
	
	xSection = new TIniFileSection(miMaxLen);
	strcpy(xSection->mcName,cSecName);
	if ( iIdx == -1 )
		iIdx = mxSections.CountItems();
	mxSections.AddItem(xSection,iIdx);
	return xSection;
}
//------------------------------------------------------------------------------
bool	TIniFile::DeleteKey(const char *cKey)
{
	bool	bDelete;
	
	bDelete = mxSecCurrent->DeletePair(cKey);
	if ( bDelete )
		mbModified = true;
	return bDelete;
}
//------------------------------------------------------------------------------
bool	TIniFile::DeleteSection(const char *cSecName)
{
	TIniFileSection	*xSecDelete;

	if ( (xSecDelete = FindSection(cSecName)) == NULL )
		return false;
	mxSections.RemoveItem(xSecDelete);
	if ( strcmp(cSecName,GlobalSection) == 0 )
		CreateSection(GlobalSection,0);
	if ( strcmp(mxSecCurrent->mcName,cSecName) == 0 )
		mxSecCurrent = (TIniFileSection *)mxSections.ItemAt(0);
	delete xSecDelete;
	mbModified = true;
	return true;
}
//------------------------------------------------------------------------------
void	TIniFile::DestroyVar()
{
	int		i;

	for ( i = mxSections.CountItems()-1 ; i >= 0 ; i-- ) {
		delete (TIniFileSection *)mxSections.ItemAt(i);
	}
	mxSections.MakeEmpty();
}
//------------------------------------------------------------------------------
void	TIniFile::EnumKeys(BList *xList)
{
	mxSecCurrent->EnumKeys(xList);
}
//------------------------------------------------------------------------------
void	TIniFile::EnumSections(BList *xList)
{
	int		i;
	
	for ( i = 0 ; i < mxSections.CountItems() ; i++ ) {
		xList->AddItem(((TIniFileSection *)mxSections.ItemAt(i))->mcName);
	}
}
//------------------------------------------------------------------------------
TIniFileSection	*TIniFile::FindSection(const char *cSecName)
{
	int				i;
	TIniFileSection	*xSectionResult = NULL;
	TIniFileSection	*xSectionTmp;
	
	for ( i = 0 ; i < mxSections.CountItems() ; i++ ) {
		xSectionTmp = (TIniFileSection *)mxSections.ItemAt(i);
		if ( BString(xSectionTmp->mcName).ToUpper() == BString(cSecName).ToUpper() ) {
			xSectionResult = xSectionTmp;
			break;
		}
	}
	return xSectionResult;
}
//------------------------------------------------------------------------------
bool	TIniFile::Flush(bool bForce)
{
	int				i,j;
	char			*cComment;
	TIniFilePair	*xCurPair;
	TIniFileSection	*xSecTmp;
	
	if ( !mbModified && !bForce)
		return false;
	Seek(0,SEEK_SET);
	SetSize(0);
	for ( i = 0 ; i < mxSections.CountItems() ; i++ ) {
		xSecTmp = (TIniFileSection *)mxSections.ItemAt(i);
		if ( strcmp(xSecTmp->mcName,GlobalSection) != 0 ) {
			Write("[ ",2);
			Write(xSecTmp->mcName,strlen(xSecTmp->mcName));
			WriteLine(" ]");
		}
		xSecTmp->miCommentPos = 0;
		cComment = new char [miMaxLen+1];
		while ( xSecTmp->ReadComment(cComment) ) {
			Write("; ",2);
			WriteLine(cComment);
		}
		delete [] cComment;
		for ( j = 0 ; j < xSecTmp->mxPairs.CountItems() ; j++ ) {
			xCurPair = (TIniFilePair *)xSecTmp->mxPairs.ItemAt(j);
			Write(xCurPair->mcKey,strlen(xCurPair->mcKey));
			Write(" = ",3);
			WriteLine(xCurPair->mcValue);
		}
		WriteLine("");
	}
	mbModified = false;
	return true;
}
//------------------------------------------------------------------------------
#if 0
int		TIniFile::GetSectionIdx(const char *cSecName)
{
	int				iIdx = -1;
	int				i;
	TIniFileSection	*xSectionTmp;
	
	for ( i = 0 ; i < mxSections.CountItems() ; i++ ) {
		xSectionTmp = (TIniFileSection *)mxSections.ItemAt(i);
		if ( BString(xSectionTmp->mxName).ToUpper() == BString(cSecName).ToUpper() ) {
			iIdx = i;
			break;
		}
	}
	return iIdx;
}
#endif
//------------------------------------------------------------------------------
void	TIniFile::Init(int iIniFileMaxLen)
{
	mbModified = false;
	miMaxLen = iIniFileMaxLen;
	Load();
	mxSecCurrent = FindSection(GlobalSection);
}		
//------------------------------------------------------------------------------
bool	TIniFile::KeyExists(const char *cKey)
{
	return (mxSecCurrent->FindPair(cKey) != NULL);
}
//------------------------------------------------------------------------------
void	TIniFile::Load()
{
	char			*cKey;
	char			*cLine;
	char			*cValue;
	int				a;
	TIniFileSection	*xSecTmp;

	//??:B_WRITE_ONLYの時は自動的に読めなくなる
	cLine = new char [miMaxLen+1];
	cKey = new char [miMaxLen+1];
	cValue = new char [miMaxLen+1];
	xSecTmp = CreateSection(GlobalSection);
	while ( !Eof() ) {
		ReadLine(cLine,miMaxLen+1);
		trimleft(cLine,cLine);
		switch ( cLine[0] ) {
		case '\0':
			continue;
		case ';':
			mid(cLine,cLine,2);
			trim(cLine,cLine);
			xSecTmp->AddComment(cLine);
			break;
		case '[':
			mid(cLine,cLine,2,instr(cLine,"]")-2);
			trim(cLine,cLine);
			xSecTmp = CreateSection(cLine);
			break;
		default:
			a = instr(cLine,"=");
			mid(cKey,cLine,1,a-1);
			trim(cKey,cKey);
			mid(cValue,cLine,a+1);
			trim(cValue,cValue);
			xSecTmp->CreatePair(cKey,cValue);
		}
	}
	delete [] cValue;
	delete [] cKey;
	delete [] cLine;		
}
//------------------------------------------------------------------------------
bool	TIniFile::ReadComment(char *cComment)
{
	return mxSecCurrent->ReadComment(cComment);
}
//------------------------------------------------------------------------------
bool	TIniFile::ReadValue(const char *cKey,char *cValue)
{
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	strcpy(cValue,xPair->mcValue);
	return true;
}
//------------------------------------------------------------------------------
bool	TIniFile::ReadValue(const char *cKey,bool *bValue)
{
	istrstream		*xSrc;
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mcValue,0);
	*xSrc >> *bValue;
	delete xSrc;
	return true;
}
//------------------------------------------------------------------------------
//#if 0
bool	TIniFile::ReadValue(const char *cKey,int32 *iValue)
{
	istrstream		*xSrc;
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mcValue,0);
	*xSrc >> *iValue;
	delete xSrc;
	return true;
}
//------------------------------------------------------------------------------
bool	TIniFile::ReadValue(const char *cKey,int64 *iValue)
{
	istrstream		*xSrc;
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mcValue,0);
	*xSrc >> *iValue;
	delete xSrc;
	return true;
}
//#endif
//------------------------------------------------------------------------------
bool	TIniFile::ReadValue(const char *cKey,float *fValue)
{
	istrstream		*xSrc;
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mcValue,0);
	*xSrc >> *fValue;
	delete xSrc;
	return true;
}
//------------------------------------------------------------------------------
bool	TIniFile::ReadValue(const char *cKey,rgb_color *xValue)
{
//int		iTmp;
//ReadValue("abc",&iTmp);

	int				iRed,iGreen,iBlue,iAlpha;
	istrstream		*xSrc;
	TIniFilePair	*xPair;

	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mcValue,0);
	*xSrc >> iRed >> iGreen >> iBlue >> iAlpha;
	xValue->red = iRed;
	xValue->green = iGreen;
	xValue->blue = iBlue;
	xValue->alpha = iAlpha;
	delete xSrc;
	return true;
}
//------------------------------------------------------------------------------
#if 0
template <class T> bool TIniFile::ReadValue(const char *cKey,T *x)
//template <class T> bool	TIniFile<T>::ReadValue(const char *cKey,T *xValue)
{
#if 0
	istrstream		*xSrc;
	TIniFilePair	*xPair;
	TIniFileSection	*xSection;
	
	if ( (xSection = FindSection(cSecName)) == NULL )
		return false;
	if ( (xPair = xSection->FindPair(cKey)) == NULL )
		return false;
	xSrc = new istrstream(xPair->mxValue.String(),0);
	*xSrc >> *xValue;
	delete xSrc;
	return true;
#endif
return true;
}
#endif
//------------------------------------------------------------------------------
bool	TIniFile::SectionExists(const char *cSecName)
{
	return (FindSection(cSecName) != NULL);
}
//------------------------------------------------------------------------------
status_t	TIniFile::SetTo(const entry_ref *ref,uint32 openMode,int iMaxLen)
{
	status_t	xResult;
	Unset();
	xResult = TSHIFile::SetTo(ref,openMode);
	Init(iMaxLen);
	return xResult;
}
//------------------------------------------------------------------------------
status_t	TIniFile::SetTo(const BEntry *entry,uint32 openMode,int iMaxLen)
{
	status_t	xResult;
	Unset();
	xResult = TSHIFile::SetTo(entry,openMode);
	Init(iMaxLen);
	return xResult;
}
//------------------------------------------------------------------------------
status_t	TIniFile::SetTo(const char *path,uint32 openMode,int iMaxLen)
{
	status_t	xResult;
	Unset();
	xResult = TSHIFile::SetTo(path,openMode);
	Init(iMaxLen);
	return xResult;
}
//------------------------------------------------------------------------------
status_t	TIniFile::SetTo(const BDirectory *dir,const char *path,uint32 openMode,int iMaxLen)
{
	status_t	xResult;
	Unset();
	xResult = TSHIFile::SetTo(dir,path,openMode);
	Init(iMaxLen);
	return xResult;
}
//------------------------------------------------------------------------------
void	TIniFile::Unset()
{
	Flush();
	DestroyVar();
	TSHIFile::Unset();
}
//------------------------------------------------------------------------------
void	TIniFile::WriteComment(const char *cComment)
{
	mxSecCurrent->AddComment(cComment);
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,const char *cValue)
{
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	strcpy(xPair->mcValue,cValue);
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,bool bValue)
{
	ostrstream		*xOut;
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	xOut = new ostrstream(xPair->mcValue,miMaxLen+1);
	*xOut << bValue;
	delete xOut;
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,int32 iValue)
{
	ostrstream		*xOut;
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	xOut = new ostrstream(xPair->mcValue,miMaxLen+1);
	*xOut << iValue;
	delete xOut;
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,int64 iValue)
{
	ostrstream		*xOut;
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	xOut = new ostrstream(xPair->mcValue,miMaxLen+1);
	*xOut << iValue;
	delete xOut;
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,float fValue)
{
	ostrstream		*xOut;
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	xOut = new ostrstream(xPair->mcValue,miMaxLen+1);
	*xOut << fValue;
	delete xOut;
	mbModified = true;
}
//------------------------------------------------------------------------------
void	TIniFile::WriteValue(const char *cKey,rgb_color xValue)
{
	char			szTmp[20];
	ostrstream		xOut(szTmp,sizeof(szTmp));
	TIniFilePair	*xPair;
	
	if ( (xPair = mxSecCurrent->FindPair(cKey)) == NULL )
		xPair = mxSecCurrent->CreatePair(cKey);
	xOut << (int)xValue.red << " " << (int)xValue.green << " " << (int)xValue.blue
			<< " " << (int)xValue.alpha << '\0';
	strcpy(xPair->mcValue,szTmp);
	mbModified = true;
#ifdef DEBUGz
	char a = szTmp[13];
	char b = szTmp[14];
	char c = szTmp[15];
#endif
}
//==============================================================================
TIniFilePair::TIniFilePair(int iMaxLen)
{
	mcKey = new char [iMaxLen+1];
	mcValue = new char [iMaxLen+1];
}
//------------------------------------------------------------------------------
TIniFilePair::~TIniFilePair()
{
	delete [] mcValue;
	delete [] mcKey;
}
//==============================================================================
TIniFileSection::TIniFileSection(int iMaxLen)
{
	miCommentPos = 0;
	miMaxLen = iMaxLen;
	mxComment = "";
	mcName = new char [miMaxLen+1];
}
//------------------------------------------------------------------------------
TIniFileSection::~TIniFileSection()
{
	int		i;
	
	for ( i = mxPairs.CountItems()-1 ; i >= 0 ; i-- ) {
		delete (TIniFilePair *)mxPairs.ItemAt(i);
	}
	delete [] mcName;
}
//------------------------------------------------------------------------------
void	TIniFileSection::AddComment(const char *cAddComment)
{
	mxComment += BString(cAddComment);
	mxComment += BString("\n");
#ifdef DEBUG
	char szTmp[1024];
	strcpy(szTmp,mxComment.String());
#endif
}
//------------------------------------------------------------------------------
TIniFilePair	*TIniFileSection::CreatePair(const char *cKey,const char *cValue)
{
	TIniFilePair	*xPair;
	
	xPair = new TIniFilePair(miMaxLen);
	strcpy(xPair->mcKey,cKey);
	strcpy(xPair->mcValue,cValue);
	mxPairs.AddItem(xPair);
	return xPair;
}
//------------------------------------------------------------------------------
bool	TIniFileSection::DeletePair(const char *cKey)
{
	TIniFilePair	*xPairTmp;

	if ( (xPairTmp = FindPair(cKey)) == NULL )
		return false;
#ifdef DEBUGz
	bool b = mxPairs.RemoveItem(xPairTmp);
#else
	mxPairs.RemoveItem(xPairTmp);
#endif
	delete xPairTmp;
	return true;
}
//------------------------------------------------------------------------------
void	TIniFileSection::EnumKeys(BList *xPair)
{
	int		i;
	
	for ( i = 0 ; i < mxPairs.CountItems() ; i++ ) {
		xPair->AddItem(((TIniFilePair *)mxPairs.ItemAt(i))->mcKey);
	}
}
//------------------------------------------------------------------------------
TIniFilePair	*TIniFileSection::FindPair(const char *cKey)
{
	int				i;
	TIniFilePair	*xPairResult = NULL;
	TIniFilePair	*xPairTmp;
	
	for ( i = 0 ; i < mxPairs.CountItems() ; i++ ) {
		xPairTmp = (TIniFilePair *)mxPairs.ItemAt(i);
		if ( BString(xPairTmp->mcKey).ToUpper() == BString(cKey).ToUpper() ) {
			xPairResult = xPairTmp;
			break;
		}
	}
	return xPairResult;
}
//------------------------------------------------------------------------------
bool	TIniFileSection::ReadComment(char *cReadComment)
{
	int		iPos = 0;
#ifdef DEBUGz
	int a = xComment.Length();
#endif
	if ( miCommentPos >= mxComment.Length() )
		return false;
//	cReadComment[0] = '\0';
	while ( mxComment[miCommentPos] != '\n' ) {
		cReadComment[iPos] = mxComment[miCommentPos];
		miCommentPos++;
		iPos++;
	}
	cReadComment[iPos] = '\0';
	miCommentPos++;
	return true;
}
//------------------------------------------------------------------------------
